/* Jungle Monkey
 * Copyright (C) 1999-2001  The Regents of the University of Michigan
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include "ggui_search.h"
#include "ggui.h"
#include "ggui_channels.h"
#include "ggui_config.h"
#include "ggui_download.h"
#include "ggui_files.h"
#include "ggui_properties.h"
#include "entry_dialog.h"
#include "pixmaps.h"

#include "jmmsearch/jmmsearch.h"
#include "jmchannel/jmchannel.h"
#include "util/util.h"
#include "util/jmgnet.h"



typedef struct _Search
{
  JMMSearch* 	search;
  GSList*    	matches;
  gchar*     	text;
  JMMSPSearchID id;
  gint	     	status;

  GURL*		channel;
  gchar*	path;

} Search;

enum { SEARCH_STATUS_NONE,
       SEARCH_STATUS_ERROR,
       SEARCH_STATUS_NOFOUND };

typedef JMMSPReply Match;

static GladeXML* 	menu_xml = NULL;
static GtkCList* 	search_clist = NULL;
static GtkCList* 	match_clist = NULL;
static GtkContainer* 	match_container = NULL;
/* TODO: match container should be "Search for" part, not whole clist */
static GtkEntry* 	search_entry = NULL;
static GtkWidget*	search_popup = NULL;

static GHashTable*	url_to_search = NULL;
static Search*		selected_search = NULL;
static Match*		selected_match = NULL;
static JMMSPVFS*	jmmsp_vfs = NULL;



static void    shutdown_hfunc (gpointer key, gpointer value, gpointer user_data);

static void    load_search_func (const ElfNode* elf);
static void    load_search_ann_func (const ElfNode* elf);
static void    save_search_hfunc (gpointer key, gpointer value, gpointer user_data);
static void    save_search_ann_hfunc (gpointer key, gpointer value, gpointer user_data);

static void    create (const GURL* channel, const gchar* path, const gchar* name);
static void    join (const GURL* url);

static Search* search_new (JMMSearch* jmmsearch);
static void    search_update (Search* search);
#define        search_get(U) ((Search*) g_hash_table_lookup(url_to_search, U))
static void    search_delete (Search* search);
static void    match_update (Match* match);

void on_search_select_row (GtkCList* list, gint row, gint column, GdkEvent* event);
void on_search_unselect_row (GtkCList* list, gint row, gint column, GdkEvent* event);
gint on_search_button_press (GtkCList* list, GdkEventButton *event, gpointer user_data);

void on_match_select_row (GtkCList* list, gint row, gint column, GdkEvent* event);
void on_match_unselect_row (GtkCList* list, gint row, gint column, GdkEvent* event);
gint on_match_button_press (GtkCList* list, GdkEventButton *event, gpointer user_data);

void on_search_entry_activate (GtkEntry* entry, gpointer user_data);

typedef enum
{
  IT_CREATE,
  IT_OPEN,
  IT_JOIN,
  IT_LEAVE,
  IT_REMOVE,
  
  IT_GET_FILE,
  IT_CANCEL_DOWNLOAD,

  IT_SEPARATOR1,
  IT_SEPARATOR2,
  IT_SEPARATOR3,
  
  IT_PROPERTIES,
  IT_PROPERTIES2,

} Item;

static void toolbar_append (Item item, GdkPixmap* gpixmap, GdkBitmap* gbitmap,
			    const gchar* tooltip, void* func);
static void update_toolbar (void);

static void update_items (void* widget, int clist); /* 0 = search, 1 = match */
static void show_item (void* widget, Item it);


#define POPUPFUNC(F) void F(gpointer callback_data, guint callback_action, GtkWidget* widget)
POPUPFUNC (on_search_get_file);
POPUPFUNC (on_search_cancel_download);
POPUPFUNC (on_search_join);
POPUPFUNC (on_search_leave);
POPUPFUNC (on_search_remove);
POPUPFUNC (on_search_create);
POPUPFUNC (on_search_open);
POPUPFUNC (on_search_properties);
POPUPFUNC (on_search_properties2);

static void search_cb (JMMSPSearch* search, JMMSPSearchID id, 
		       const JMMSPReply* reply, gpointer user_data);
static void jmmsp_error_cb (JMMSP* jmmsp, gpointer user_data);



void
ggui_search_init (void)
{
  menu_xml = ggui_get_glade_xml ("search_popup", "master_popup.glade");
  g_assert (menu_xml != NULL);
  glade_xml_signal_autoconnect (menu_xml);

  search_clist = GTK_CLIST(glade_xml_get_widget(jmw_xml, "jmsearchclist"));
  search_entry = GTK_ENTRY(glade_xml_get_widget(jmw_xml, "jmsearchentry"));
  match_clist  = GTK_CLIST(glade_xml_get_widget(jmw_xml, "jmmatchclist"));
  match_container = GTK_CONTAINER(glade_xml_get_widget(jmw_xml, "search_container"));
  search_popup = GTK_WIDGET(glade_xml_get_widget(menu_xml, "search_popup"));

  g_assert (search_clist != NULL);
  g_assert (search_entry != NULL);
  g_assert (match_clist  != NULL);
  g_assert (match_container  != NULL);
  g_assert (search_popup != NULL);

  gtk_clist_column_title_passive (match_clist, 0);
  gtk_clist_column_title_passive (match_clist, 1);

  gtk_widget_set_sensitive (GTK_WIDGET(match_container), FALSE);

  ggui_set_tooltip (GTK_WIDGET(search_entry), 
		    _("Seach query.  "
		      "Enter a string to search for and press return.  "  
		      "Regular expressions are _not_ supported yet."));
  ggui_set_tooltip (GTK_WIDGET(match_clist), 
		    _("Results from the last search."
		      "Double click to download."));
  /* TODO: Complete tooltips */

  url_to_search = g_hash_table_new (gnet_url_hash, gnet_url_equal);
  jmmsp_vfs = jmmsp_vfs_new ();
}


void
ggui_search_shutdown (void)
{
  /* Delete searches */
  g_hash_table_foreach (url_to_search, shutdown_hfunc, NULL);
  g_hash_table_destroy (url_to_search);
  url_to_search = NULL;
  selected_search = NULL;
  selected_match = NULL;

  /* Delete VFS */
  jmmsp_vfs_delete (jmmsp_vfs);
  jmmsp_vfs = NULL;
}


static void
shutdown_hfunc (gpointer key, gpointer value, gpointer user_data)
{
  Search* search = (Search*) value;
  search_delete (search);
}


void
ggui_search_load (void)
{
  ggui_load_elf ("search",          load_search_func);
  ggui_load_elf ("search-announce", load_search_ann_func);
}



static void
load_search_func (const ElfNode* elf)
{
  const gchar* name;
  const gchar* url_str;
  GURL*   url = NULL;
  JMMSearch* jmmsearch = NULL;

  name    = elf_get_attribute (elf, "name");
  url_str = elf_get_attribute (elf, "url");
  if (url_str) url = gnet_url_new (url_str);

  if (!strcmp (elf->name, "create") && name)
    jmmsearch = jmmsearch_create_full (jm_btp_bpeer, name, NULL, NULL, jmmsp_vfs);
  else if (!strcmp (elf->name, "join") && url && !search_get(url))
    jmmsearch = jmmsearch_join_full (jm_btp_bpeer, url, NULL, NULL, jmmsp_vfs);
  else if (!strcmp (elf->name, "setup") && url && !search_get(url))
    jmmsearch = jmmsearch_setup_full (jm_btp_bpeer, url, NULL, NULL, jmmsp_vfs);
  if  (jmmsearch)
    search_new (jmmsearch);

  gnet_url_delete (url);
}


static void
load_search_ann_func (const ElfNode* elf)
{
  const gchar*  curl_str;
  const gchar*  url_str;
  const gchar*  path;
  GURL*   curl = NULL;
  GURL*   url  = NULL;
  Search* search;
  JMAnn*  ann;

  curl_str = elf_get_attribute (elf, "curl");
  url_str  = elf_get_attribute (elf, "url");
  path     = elf_get_attribute (elf, "path");
  if (!curl_str || !url_str || !path) return;

  curl = gnet_url_new (curl_str);
  url = gnet_url_new (url_str);
  if (!url || !curl) goto done;
      
  /* Change URL's */
  if (!SAFESTRCMP(jm_hostname, curl->hostname) &&
      (!jm_btp_last_port || curl->port == jm_btp_last_port))
    gnet_url_set_port (curl, jm_btp_bpeer->port);
  if (!SAFESTRCMP(jm_hostname, url->hostname) &&
      (!jm_btp_last_port || url->port == jm_btp_last_port))
    gnet_url_set_port (url, jm_btp_bpeer->port);

  search = search_get (url);
  if (!search || search->channel) goto done; /* TODO: Multiple paths */

  ann = ggui_channel_add_path (curl, path, url);
  if (!ann) goto done;

  search->channel = curl; curl = NULL;
  search->path    = g_strdup (ann->path);

 done:
  gnet_url_delete (url);
  gnet_url_delete (curl);

}


void
ggui_search_save (void)
{
  ggui_save_hashtable ("search",          url_to_search, save_search_hfunc);
  ggui_save_hashtable ("search-announce", url_to_search, save_search_ann_hfunc);
}


static void
save_search_hfunc (gpointer key, gpointer value, gpointer user_data)
{
  Search*  search = (Search*) value;
  FILE*    file = (FILE*) user_data;
  ElfNode* elf;
  gchar*   buf;
  gint     len;

  elf = elf_new (NULL);

  if (search->search->jmmsp->is_local)
    {
      elf_set_name (elf, "create");
      elf_set_attribute (elf, "name", gnet_url_basename (search->search->jmmsp->url));
    }
  else if (jmmsearch_is_up (search->search))
    {
      gchar* url_str;

      elf_set_name (elf, "join");
      url_str = gnet_url_get_nice_string (search->search->jmmsp->url);
      elf_set_attribute (elf, "url", url_str);
      g_free (url_str);
    }
  else
    {
      gchar* url_str;

      elf_set_name (elf, "setup");
      url_str = gnet_url_get_nice_string (search->search->jmmsp->url);
      elf_set_attribute (elf, "url", url_str);
      g_free (url_str);
    }

  elf_write (elf, &buf, &len);
  fwrite (buf, len, 1, file);

  elf_delete (elf);
  g_free (buf);
}


static void
save_search_ann_hfunc (gpointer key, gpointer value, gpointer user_data)
{
  Search*  search = (Search*) value;
  FILE*    file = (FILE*) user_data;
  gchar*   curl;
  gchar*   url;
  ElfNode* elf;
  gchar*   buf;
  gint     len;

  if (!(search->channel && search->path))
    return;

  url  = gnet_url_get_nice_string (search->search->jmmsp->url);
  curl = gnet_url_get_nice_string (search->channel);

  elf = elf_new_format ("ann", "sss", "curl", curl, "path", search->path, 
			"url", url);
  elf_write (elf, &buf, &len);
  fwrite (buf, len, 1, file);

  elf_delete (elf);
  g_free (buf);
  g_free (url);
  g_free (curl);
}





/* ******************** */

JMMSearch*  
ggui_search_get (const GURL* url)
{
  Search* search;

  g_return_val_if_fail (url, NULL);

  search = (Search*) g_hash_table_lookup (url_to_search, url);
  if (!search)
    return NULL;

  return search->search;
}


gboolean
ggui_search_is_up (const GURL* url)
{
  JMMSearch* jmmsearch;

  g_return_val_if_fail (url, FALSE);

  jmmsearch = ggui_search_get (url);
  if (!jmmsearch)
    return FALSE;

  return jmmsearch_is_up (jmmsearch);
}


gboolean
ggui_search_is_down (const GURL* url)
{
  JMMSearch* jmmsearch;

  g_return_val_if_fail (url, FALSE);

  jmmsearch = ggui_search_get (url);
  if (!jmmsearch)
    return FALSE;

  return jmmsearch_is_down (jmmsearch);
}


gchar*   
ggui_search_get_properties (const GURL* url)
{
  JMMSearch* jmmsearch;

  jmmsearch = ggui_search_get (url);
  if (!jmmsearch)
    return g_strdup ("");

  return g_strdup_printf (_("Search %s\n"), gnet_url_basename(url));
}



/* ******************** */

static EntryDialog* create_search_dialog  = NULL;
static GURL*        create_search_channel = NULL;
static gchar* 	    create_search_path    = NULL;

static void	    create_ok (gchar* search_name, gchar* ignore);
static void	    create_cancel (void);

void
ggui_search_create (const GURL* channel, const gchar* path, const gchar* name)
{
  g_return_if_fail ((channel && path) || !(channel && path));

  /* Prompt for name if not given */
  if (!name)
    {
      if (create_search_dialog)
	{
	  entry_dialog_show (create_search_dialog);
	  return;
	}

      if (channel && path)
	{
	  create_search_channel = gnet_url_clone (channel);
	  create_search_path = g_strdup (path);
	}

      create_search_dialog = entry_dialog_new(_("Create search"), 
					      _("Search name"), NULL,
					      create_ok, create_cancel);
      return;
    }

  create (channel, path, name);
}


static void
create_ok (gchar* search_name, gchar* ignore)
{
  create (create_search_channel, create_search_path, search_name);
  create_cancel ();
}


static void
create_cancel (void)
{
  create_search_dialog = NULL;
  gnet_url_delete (create_search_channel);
  create_search_channel = NULL;
  g_free (create_search_path);
  create_search_path = NULL;
}


static void
create (const GURL* channel, const gchar* path, const gchar* name)
{
  JMMSearch*    jmmsearch;
  Search*	search;

  g_return_if_fail (name);
  g_return_if_fail ((channel && path) || !(channel && path));

  /* Check if path ok */
  if (channel && path && !ggui_channel_check_path (channel, path, name))
    return;

  /* Create the channel */
  jmmsearch = jmmsearch_create_full (jm_btp_bpeer, name, NULL, NULL, jmmsp_vfs);
  if (!jmmsearch)
    {
      ggui_show_message ("warning", _("Could not create search \"%s\"\n"), name);
      return;
    }
  
  /* Add it */
  search = search_new (jmmsearch);
  g_return_if_fail (search);

  /* Announce it */
  if (channel && path)
    {
      JMAnn* ann;

      ann = ggui_channel_add (channel, path, name, jmmsearch->jmmsp->url);
      g_return_if_fail (ann);
      search->channel = gnet_url_clone (channel);
      search->path = g_strdup (ann->path);
    }
}


/* ********** */

static EntryDialog* join_dialog  = NULL;

static void join_ok (gchar* url, gchar* ignore);
static void join_cancel (void);


void
ggui_search_join (const GURL* url)
{
  Search* search;

  /* Prompt for name if not given */
  if (!url)
    {
      if (join_dialog)
	{
	  entry_dialog_show (join_dialog);
	  return;
	}

      join_dialog = entry_dialog_new(_("Open search"), 
				     _("Search URL"), NULL,
				     join_ok, join_cancel);

      return;
    }

  /* Check if search exists */
  search = (Search*) g_hash_table_lookup (url_to_search, url);
  if (search)
    {
      if (!jmmsearch_is_up (search->search))
	{
	  jmmsearch_rejoin (search->search);
	  search_update (search);
	}
      return;
    }

  /* Create it */
  join (url);
}


static void
join_ok (gchar* url, gchar* ignore)
{
  GURL* gurl;

  gurl = gnet_url_new (url);
  if (gurl && gnet_url_is_scheme(gurl, "jmmsp"))
    {
      join (gurl);
      gnet_url_delete (gurl);
    }
  else
    {
      ggui_show_message ("warning", _("Malformed URL: %s  (should be jmmsp://something)\n"));
    }

  join_cancel ();
}


static void
join_cancel(void)
{
  join_dialog = NULL;
}


static void
join (const GURL* url)
{
  Search*   search;
  JMMSearch* jmmsearch;

  g_return_if_fail (url);

  /* Check if search exists */
  search = (Search*) g_hash_table_lookup (url_to_search, url);
  if (search)
    {
      if (!jmmsearch_is_up (search->search))
	{
	  jmmsearch_rejoin (search->search);
	  search_update (search);
	}

      ggui_show_message ("warning", _("Search is already open\n"));
      return;
    }

  /* Create search */
  jmmsearch = jmmsearch_join_full (jm_btp_bpeer, url, NULL, NULL, jmmsp_vfs);
  if (!jmmsearch)
    {
      ggui_show_message ("warning", _("Could not open search\n"));
      return;
    }

  /* Add it */
  search = search_new (jmmsearch);
}


/* ********** */

static Search*
search_new (JMMSearch* jmmsearch)
{
  Search*    search;
  gchar*     texts[2] = {NULL, NULL};
  gint 	     row;

  g_return_val_if_fail (jmmsearch, NULL);
  g_return_val_if_fail (jmmsearch->jmmsp->url, NULL);

  /* Make sure search isn't there */
  search = (Search*) g_hash_table_lookup (url_to_search, jmmsearch->jmmsp->url);
  g_return_val_if_fail (!search, NULL);
  
  /* Create new search */
  search = g_new0 (Search, 1);
  search->search = jmmsearch;
  g_hash_table_insert (url_to_search, jmmsearch->jmmsp->url, search);

  /* Initialize it */
  jmmsearch->jmmsp->error_func = jmmsp_error_cb;
  jmmsearch->jmmsp->user_data  = search;

  /* Add to clist */
  texts[0] = gnet_url_basename(jmmsearch->jmmsp->url);
  row = gtk_clist_append (search_clist, texts);
  gtk_clist_set_row_data (search_clist, row, search);

  /* Update */
  search_update (search);

  /* Select it if no other is selected */
  if (!ggui_search_get_selected())
    ggui_search_select (jmmsearch->jmmsp->url);

  return search;
}


static void
search_update (Search* search)
{
  GdkPixmap* pixmap;
  GdkBitmap* mask;
  gint	     row;
  GURL*	     url;

  g_return_if_fail (search);
  url = search->search->jmmsp->url;

  /* Update icons */
  row = gtk_clist_find_row_from_data (search_clist, search);
  ggui_get_pixmaps (url, &pixmap, &mask);
  gtk_clist_set_pixtext (search_clist, row, 0, gnet_url_basename(url), 5, pixmap, mask);

  /* Set sensitivity and toolbar */
  if (search == selected_search)
    {
      update_toolbar ();
      gtk_widget_set_sensitive (GTK_WIDGET(match_container), 
				jmmsearch_is_up(search->search));
    }

  /* Update files */
  ggui_file_update (url);
}


static void
search_delete (Search* search)
{
  GSList* i;

  jmmsearch_leave (search->search);
  if (search->id)
    jmmsp_search_cancel (search->search->jmmsp_search, search->id);
  search->search->jmmsp_vfs = NULL;
  jmmsp_remove_handler (search->search->jmmsp, (JMMSPHandler*) jmmsp_vfs);
  jmmsearch_delete (search->search);
  for (i = search->matches; i != NULL; i = i->next)
    {
      Match* match = (Match*) i->data;
      jmmsp_reply_delete (match);
    }
  g_slist_free (search->matches);
  g_free (search->text);
  gnet_url_delete (search->channel);
  g_free (search->path);
  g_free (search); 
}


static void
match_update (Match* match)
{
  gint	     row;
  GdkPixmap* pixmap;
  GdkBitmap* mask;
  gint	     percent;
  gchar*     name;
  gchar*     text1 = NULL;
  gchar*     text2 = NULL;

  /* Update only if in match clist */
  row = gtk_clist_find_row_from_data (match_clist, match);
  if (row < 0) return;

  name = gnet_url_basename (match->url);
  percent = ggui_download_get_percent_complete (match->url);
  if (percent >= 0 && percent < 100)
    text1 = g_strdup_printf (_("%s (downloading)"), name);
  else if (percent == -2)
    text1 = g_strdup_printf (_("%s (down)"), name);
  else
    text1 = g_strdup (name);

  text2 = gnet_url_get_nice_string (match->url);

  ggui_get_pixmaps (match->url, &pixmap, &mask);
  ggui_clist_update_row2 (match_clist, row, text1, text2, pixmap, mask);
  g_free (text1);
  g_free (text2);
}




/* ********** */


void
ggui_search_leave (const GURL* url)
{
  Search* search;
  GSList* i;

  g_return_if_fail (url);

  /* Get search */
  search = (Search*) g_hash_table_lookup (url_to_search, url);
  if (!search)
    return;

  /* Leave group */
  jmmsearch_leave (search->search);
  g_free (search->text);
  search->text = NULL;
  if (search->id)
    {
      jmmsp_search_cancel (search->search->jmmsp_search, search->id);
      search->id = NULL;
    }
  for (i = search->matches; i != NULL; )
    {
      Match* match = (Match*) i->data;
      i = i->next;

      jmmsp_reply_delete (match);
    }
  g_slist_free (search->matches);
  search->matches = NULL;

  /* Unselect search */
  if (search == selected_search)
    {
      gtk_clist_clear (match_clist);
      gtk_widget_set_sensitive (GTK_WIDGET(match_container), FALSE);
    }

  /* Unselect all matches */
  selected_match = NULL;
  gtk_clist_unselect_all (match_clist);

  search_update (search);
}


void
ggui_search_remove (const GURL* url)
{
  Search* search;
  JMMSP*  jmmsp;
  gint  row;

  g_return_if_fail (url);

  /* Get search */
  search = (Search*) g_hash_table_lookup (url_to_search, url);
  if (!search)
    return;
  g_hash_table_remove (url_to_search, url);
  jmmsp = search->search->jmmsp;

  /* Remove the search from the list (this will unselect the search if it
     is selected) */
  row = gtk_clist_find_row_from_data (search_clist, search);
  if (row >= 0)
    gtk_clist_remove (search_clist, row);

  /* Send remove */
  if (jmmsp->is_local && search->channel && search->path)
    ggui_channel_remove_path (search->channel, search->path);
  /* remove_path may delete url */

  /* Delete group */
  search_delete (search);
}


void
ggui_search_update (const GURL* url)
{
  GSList* i;

  g_return_if_fail (url);

  if (!selected_search)
    return;

  /* Check for match to update */
  for (i = selected_search->matches; i != NULL; i = i->next)
    {
      Match* m = (Match*) i->data;
      if (gnet_url_equal (m->url, url))
	match_update (m);
    }
}


void
ggui_search_select (const GURL* url)
{
  Search* search;
  gint  row;

  g_return_if_fail (url);

  /* Get search */
  search = (Search*) g_hash_table_lookup (url_to_search, url);
  if (!search)
    return;

  row = gtk_clist_find_row_from_data (search_clist, search);
  if (row >= 0)
    gtk_clist_select_row (search_clist, row, 0);
}


GURL*
ggui_search_get_selected (void)
{
  if (selected_search)
    return selected_search->search->jmmsp->url;

  return NULL;
}


JMMSPVFSID*  
ggui_search_vfs_add (const GURL* url, const gchar* name, gint length)
{
  return jmmsp_vfs_add (jmmsp_vfs, name, url, length);
}


void
ggui_search_vfs_remove (JMMSPVFSID* vfsid)
{
  jmmsp_vfs_remove (jmmsp_vfs, vfsid);
}


/* ********** */

static void
jmmsp_error_cb (JMMSP* jmmsp, gpointer user_data)
{
  Search* search = (Search*) user_data;

  search_update (search);
}




/* **************************************** */

void
on_search_select_row (GtkCList* list, gint row, gint column, GdkEvent* event)
{
  Search* new_selected_search;

  new_selected_search = (Search*) gtk_clist_get_row_data(list, row);
  if (new_selected_search == selected_search)
    return;

  /* Save the text entry of the old selected search */
  if (selected_search)
    {
      gchar* text;

      text = gtk_entry_get_text (search_entry);
      if (text)
	text = g_strdup (text);
      
      if (selected_search->text)
	g_free (selected_search->text);

      selected_search->text = text;
    }

  /* Restore the text entry of the new selected sarch */
  if (new_selected_search)
    {
      gchar* text;

      gtk_entry_set_text (search_entry, "");

      /* Restore text */
      text = new_selected_search->text;
      if (text)
	gtk_entry_set_text (search_entry, text);

      /* Show matches */
      gtk_clist_clear (match_clist);
      if (new_selected_search->matches)
	{
	  GSList* i;

	  gtk_clist_freeze (match_clist);

	  for (i = new_selected_search->matches; i != NULL; i = i->next)
	    {
	      Match* match = (Match*) i->data;
	      gchar* texts[2] = {NULL, NULL};
	      int row;

	      /* Add row (blank for now) */
	      row = gtk_clist_insert (match_clist, 0, texts);
	      gtk_clist_set_row_data (match_clist, row, match);

	      /* Fill in that row */
	      match_update (match);
	    }

	  gtk_clist_thaw (match_clist);
	}
      else if (new_selected_search->id)
	{
	  const gchar* keyword;
	  gchar* texts[2] = {NULL, NULL};

	  keyword = jmmsp_search_get_keyword (new_selected_search->id);
	  g_return_if_fail (keyword);

	  texts[0] = g_strdup_printf (_("Searching for \"%s\"..."), keyword);
	  gtk_clist_insert (match_clist, 0, texts);
	  g_free (texts[0]);
	}
      else if (new_selected_search->status)
	{
	  if (new_selected_search->status == SEARCH_STATUS_ERROR)
	    {
	      gchar* texts[3] = {"Search error", "", NULL};

	      gtk_clist_clear (match_clist);
	      gtk_clist_insert (match_clist, 0, texts);
	    }
	  else if (new_selected_search->status == SEARCH_STATUS_NOFOUND)
	    {
	      gchar* texts[3] = {"No matches found", "", NULL};

	      gtk_clist_clear (match_clist);
	      gtk_clist_insert (match_clist, 0, texts);
	    }

	  /* TODO: Any error status got lost.  This should be fixed. */
	}
    }

  selected_search = new_selected_search;

  gtk_widget_set_sensitive (GTK_WIDGET(match_container), 
			    (selected_search != NULL && 
			     jmmsearch_is_up (selected_search->search)));
  update_toolbar ();
}


void
on_search_unselect_row (GtkCList* list, gint row, gint column, GdkEvent* event)
{
  if (selected_search)
    {
      gchar* text;

      /* Save text */
      text = gtk_entry_get_text (search_entry);
      if (text)
	text = g_strdup (text);
      
      if (selected_search->text)
	g_free (selected_search->text);

      selected_search->text = text;

      /* Clear matches */
      gtk_clist_clear (match_clist);
    }

  selected_search = NULL;
  gtk_widget_set_sensitive (GTK_WIDGET(match_container), FALSE);
  update_toolbar();
}


gint
on_search_button_press (GtkCList* list, GdkEventButton *event, gpointer user_data)
{
  gint row, column;
  gboolean rv = FALSE;


  /* User double clicked - activate the thing selected */
  if (event->type == GDK_2BUTTON_PRESS && 
      (event->button == 1 || event->button == 2))
    {
      /* Get the selection information */
      if (gtk_clist_get_selection_info(list, event->x, event->y, &row, &column))
	{
	  /* Select the clicked row, if there is one */
	  gtk_clist_select_row (list, row, 0);
	
	  /* Double clicks on searches. */
	  if (selected_search)
	    ggui_search_join (selected_search->search->jmmsp->url);
	}
      else
	gtk_clist_unselect_all(list);

      rv = TRUE;
    }

  /* User right-clicked - show context menu */
  else if (event->type == GDK_BUTTON_PRESS && event->button == 3)
    {
      GList* i;

      /* Select the clicked row, if there is one */
      if (gtk_clist_get_selection_info (list, event->x, event->y, &row, &column))
	gtk_clist_select_row (list, row, 0);
      else
	gtk_clist_unselect_all (list);

      /* Hide everything */
      for (i = GTK_MENU_SHELL(search_popup)->children; i != NULL; i = i->next)
	{
	  GtkWidget* menu_item = GTK_WIDGET(i->data);
	  g_assert (GTK_MENU_ITEM(menu_item) != NULL);
	  gtk_widget_hide(menu_item);
	}

      /* Update the popup menu */
      update_items (NULL, 0);

      /* Pop up */
      gtk_menu_popup (GTK_MENU(search_popup), NULL, NULL, 
		      NULL, NULL, 3, event->time);

      rv = TRUE;
    }

  return rv;
}


void 
on_match_select_row (GtkCList* list, gint row, gint column, GdkEvent* event)
{
  selected_match = (Match*) gtk_clist_get_row_data (list, row);
}


void 
on_match_unselect_row (GtkCList* list, gint row, gint column, GdkEvent* event)
{
  selected_match = NULL;
}



/* Return TRUE if we handle event; FALSE to pass on */
gint
on_match_button_press (GtkCList* list, GdkEventButton *event, gpointer user_data)
{
  gint row, column;
  gboolean rv = FALSE;

  /* User double clicked - activate the thing selected */
  if (event->type == GDK_2BUTTON_PRESS && 
      (event->button == 1 || event->button == 2))
    {
      /* Get the selection information */
      if (gtk_clist_get_selection_info(list, event->x, event->y, &row, &column))
	{
	  /* Select the clicked row, if there is one */
	  gtk_clist_select_row (list, row, 0);
	
	  /* Double clicks on matches activates them. */
	  if (selected_match)
	    ggui_activate (selected_match->url);
	}
      else
	gtk_clist_unselect_all (list);

      rv = TRUE;
    }

  /* User right-clicked - show context menu */
  else if (event->type == GDK_BUTTON_PRESS && event->button == 3)
    {
      GList* i;

      /* Select the clicked row, if there is one */
      if (gtk_clist_get_selection_info (list, event->x, event->y, &row, &column))
	gtk_clist_select_row (list, row, 0);
      else
	gtk_clist_unselect_all (list);

      /* Hide everything */
      for (i = GTK_MENU_SHELL(search_popup)->children; i != NULL; i = i->next)
	{
	  GtkWidget* menu_item = GTK_WIDGET(i->data);
	  g_assert (GTK_MENU_ITEM(menu_item) != NULL);
	  gtk_widget_hide(menu_item);
	}

      /* Build popup */
      update_items (NULL, 1);

      /* Show popup */
      gtk_menu_popup (GTK_MENU(search_popup), NULL, NULL, NULL, 
		      NULL, 3, event->time);

      rv = TRUE;
    }

  return rv;
}


void
ggui_search_show_toolbar (void)
{
  /* Add everything to the toolbar */
  toolbar_append (IT_OPEN, search_pixmap, search_mask, 
		  _("Open search"), on_search_open);
  toolbar_append (IT_CREATE, create_search_pixmap, create_search_mask,
		  _("Create search"), on_search_create);
  toolbar_append (IT_JOIN, join_search_pixmap, join_search_mask, 
		  _("Join search"), on_search_join);
  toolbar_append (IT_LEAVE, leave_search_pixmap, leave_search_mask, 
		  _("Leave search"), on_search_leave);
  /* FIX: Add X */

  toolbar_append (IT_GET_FILE, page_pixmap, page_mask, 
		  _("Get file"), on_search_get_file);
  toolbar_append (IT_CANCEL_DOWNLOAD, down_page_pixmap, down_page_mask, 
		  _("Cancel download"), on_search_cancel_download);

  /* Update the toolbar */
  update_toolbar();
}


static void
toolbar_append (Item item, GdkPixmap* gpixmap, GdkBitmap* gbitmap,
	       const gchar* tooltip, void* func)
{
  GtkPixmap* pixmap;				     	 
  GtkWidget* widget;				     	 

  pixmap = (GtkPixmap*) gtk_pixmap_new (gpixmap, gbitmap);
  g_return_if_fail (pixmap);                         	 
  widget = gtk_toolbar_append_item (jmw_toolbar, NULL,   
				    tooltip, "private",                 
				    GTK_WIDGET(pixmap),           
				    (GtkSignalFunc) func, NULL);     
  g_return_if_fail (widget);                             
  gtk_object_set_user_data (GTK_OBJECT(widget), (gpointer) item);
}


static void
update_toolbar (void)
{
  GList* kids;
  GList* i;

  /* Hide all */
  kids = gtk_container_children (GTK_CONTAINER(jmw_toolbar));
  for (i = kids; i != NULL; i = i->next)
    gtk_widget_set_sensitive (GTK_WIDGET(i->data), FALSE);

  /* Update toolbar */
  update_items (kids, 2);

  g_list_free (kids);
}


/* **************************************** */

static const char* item_mi_name[] =
{
  "create_mi",
  "open_mi",
  "join_mi",
  "leave_mi",
  "remove_mi",

  "search_get_file_mi",
  "search_cancel_download_mi",

  "separator1",
  "separator2",
  "separator3",

  "properties_mi",
  "properties2_mi",
};

static void
update_items (void* widget, int clist)
{

  show_item (widget, IT_OPEN);
  show_item (widget, IT_CREATE);

  if (selected_search)
    {
      JMMSP* jmmsp = selected_search->search->jmmsp;

      show_item (widget, IT_SEPARATOR1);

      if (jmmsp->is_local)
	show_item (widget, IT_REMOVE);
      else if (jmmsp_is_up (jmmsp))
	{
	  show_item (widget, IT_LEAVE);
	  show_item (widget, IT_REMOVE);
	}
      else
	{
	  show_item (widget, IT_JOIN);
	  show_item (widget, IT_REMOVE);
	}

      show_item (widget, IT_SEPARATOR2);

      if (!(selected_match && clist))
	{
	  show_item (widget, IT_SEPARATOR3);
	  show_item (widget, IT_PROPERTIES);
	}
    }

  if (selected_match && clist)
    {
      gint percent;

      if (ggui_file_have (selected_match->url) ||
	  (percent = ggui_download_get_percent_complete (selected_match->url)) == 100)
	/* ignore */;
      else if (percent == -1 || percent == -2)
	show_item (widget, IT_GET_FILE);
      else
	show_item (widget, IT_CANCEL_DOWNLOAD);

      show_item (widget, IT_SEPARATOR1);
      show_item (widget, IT_PROPERTIES2);
    }
}


static void
show_item (void* widget, Item it)
{
  if (!widget)	/* Popup menu */
    {
      GtkWidget* menu_item = glade_xml_get_widget(menu_xml, item_mi_name[it]);
      g_return_if_fail (menu_item != NULL);
      gtk_widget_show(menu_item);
    }
  else		/* Toolbar    */
    {
      GList* kids = (GList*) widget;
      GList* i;

      for (i = kids; i != NULL; i = i->next)		 
	{
	  if (it == (guint) gtk_object_get_user_data(GTK_OBJECT(i->data)))
	    {                                                  
	      gtk_widget_set_sensitive (GTK_WIDGET(i->data), TRUE); 
	      break;                                           
	    }                                                  
	}
    }
}



POPUPFUNC (on_search_get_file)
{
  g_return_if_fail (selected_match);
  ggui_download_get (selected_match->url, NULL, NULL);
}

POPUPFUNC (on_search_cancel_download)
{
  g_return_if_fail (selected_match);
  ggui_download_cancel (selected_match->url);
}

POPUPFUNC (on_search_join)
{
  g_return_if_fail (selected_search);
  ggui_search_join (selected_search->search->jmmsp->url);
}

POPUPFUNC (on_search_leave)
{
  g_return_if_fail (selected_search);
  ggui_search_leave (selected_search->search->jmmsp->url);
}

POPUPFUNC (on_search_remove)
{
  GURL* url;

  g_return_if_fail (selected_search);
  url = selected_search->search->jmmsp->url;

  if (jmmsearch_is_local(selected_search->search))
    {
      ggui_search_remove (url);
    }
  else if (jmmsearch_is_up (selected_search->search))
    {
      ggui_search_leave (url);
      ggui_search_remove (url);
    }
  else
    {
      ggui_search_remove (url);
    }
}

POPUPFUNC (on_search_create)
{
  ggui_search_create (NULL, NULL, NULL);
}

POPUPFUNC (on_search_open)
{
  ggui_search_join (NULL);
}

POPUPFUNC(on_search_properties)
{
  g_return_if_fail (selected_search);
  ggui_properties_show (selected_search->search->jmmsp->url);
}

POPUPFUNC(on_search_properties2)
{
  g_return_if_fail (selected_match);
  ggui_properties_show (selected_match->url);
}


/* **************************************** */

void
on_search_entry_activate (GtkEntry* entry, gpointer user_data)
{
  JMMSearch* jmmsearch;
  gchar* keyword;
  GSList* i;
  gchar* texts[2] = {NULL, NULL};

  g_return_if_fail (selected_search);
  jmmsearch = selected_search->search;
  g_return_if_fail (jmmsearch);
  g_return_if_fail (jmmsearch_is_up(jmmsearch));

  /* Get keyword */
  keyword = gtk_entry_get_text (search_entry);
  if (!keyword || !*keyword)
    return;

  /* Unlink old search results */
  for (i = selected_search->matches; i != NULL; )
    {
      Match* match = (Match*) i->data;
      i = i->next;

      jmmsp_reply_delete (match);
    }
  g_slist_free (selected_search->matches);
  selected_search->matches = NULL;

  /* Clear list of old search results */
  gtk_clist_clear (match_clist);

  /* Reset our status */
  selected_search->status = SEARCH_STATUS_NONE;

  /* Cancel any pending seach */
  if (selected_search->id)
    jmmsp_search_cancel (jmmsearch->jmmsp_search, selected_search->id);


  /* Note we are searching */
  texts[0] = g_strdup_printf (_("Searching for \"%s\"..."), keyword);
  gtk_clist_insert (match_clist, 0, texts);
  g_free (texts[0]);

  /* Reset entry */
  gtk_entry_set_text (search_entry, "");

  /* Search */
  selected_search->id = 
    jmmsp_search_query (jmmsearch->jmmsp_search, keyword, search_cb, selected_search);

  /* Note if error */
  if (!selected_search->id)
    {
      texts[0] = g_strdup_printf (_("Search error"));
      selected_search->status = SEARCH_STATUS_ERROR;
      gtk_clist_insert (match_clist, 0, texts);
      g_free (texts[0]);
    }
}



static void
search_cb (JMMSPSearch* jmmsp_search, JMMSPSearchID id, 
	   const JMMSPReply* reply, gpointer user_data)
{
  Search* search = (Search*) user_data;
  Match*  match;
  GSList* i;

  /* Check if error (or just a timeout) */
  if (!reply)
    {
      search->id = NULL;
      if (!search->matches) /* No error if there were already matches */
	{
	  search->status = SEARCH_STATUS_NOFOUND;
	  if (search == selected_search)
	    {
	      gchar* texts[3] = {"No matches found", "", NULL};

	      gtk_clist_clear (match_clist);
	      gtk_clist_insert (match_clist, 0, texts);
	    }
	}
      return;
    }
  g_return_if_fail (reply->url);

  /* Clear match clist if this is the first match */
  if (search == selected_search && !search->matches)
    gtk_clist_clear (match_clist);

  /* Ignore if we have this match */
  for (i = search->matches; i != NULL; i = i->next)
    {
      Match* m = (Match*) i->data;
      if (gnet_url_equal (m->url, reply->url))
	return;
    }

  /* Create new object */
  match = jmmsp_reply_clone (reply);
  search->matches = g_slist_prepend (search->matches, match);

  /* Add to GUI if showing */
  if (selected_search == search)
    {
      gchar* texts[2] = {NULL, NULL};
      gint row;
      
      /* Add row (blank for now) */
      row = gtk_clist_insert (match_clist, 0, texts);
      gtk_clist_set_row_data (match_clist, row, match);

      /* Fill in that row */
      match_update (match);
    }
}
